import torch
from torch import nn

import dgl.function as fn
from dgl.nn.functional import edge_softmax
from dgl.utils import expand_as_pair

from tools.utils import create_activation


class DotGAT(nn.Module):
    def __init__(self,
                 in_dim,
                 num_hidden,
                 out_dim,
                 num_layers,
                 nhead,
                 nhead_out,
                 activation,
                 feat_drop,
                 attn_drop,
                 residual,
                 norm,
                 concat_out=False,
                 encoding=False
                 ):
        super(DotGAT, self).__init__()
        self.out_dim = out_dim
        self.num_heads = nhead
        self.num_layers = num_layers
        self.gat_layers = nn.ModuleList()
        self.activation = activation
        self.concat_out = concat_out

        last_activation = create_activation(activation) if encoding else None
        last_residual = (encoding and residual)
        last_norm = norm if encoding else None
        
        if num_layers == 1:
            self.gat_layers.append(DotGatConv(
                in_dim, out_dim, nhead_out,
                feat_drop, attn_drop, last_residual, norm=last_norm, concat_out=concat_out))
        else:
            # input projection (no residual)
            self.gat_layers.append(DotGatConv(
                in_dim, num_hidden, nhead,
                feat_drop, attn_drop, residual, create_activation(activation), norm=norm, concat_out=concat_out))
            # hidden layers
            for l in range(1, num_layers - 1):
                # due to multi-head, the in_dim = num_hidden * num_heads
                self.gat_layers.append(DotGatConv(
                    num_hidden * nhead, num_hidden, nhead,
                    feat_drop, attn_drop, residual, create_activation(activation), norm=norm, concat_out=concat_out))
            # output projection
            self.gat_layers.append(DotGatConv(
                num_hidden * nhead, out_dim, nhead_out,
                feat_drop, attn_drop, last_residual, activation=last_activation, norm=last_norm, concat_out=concat_out))
    
        self.head = nn.Identity()
    
    def forward(self, g, inputs, return_hidden=False):
        h = inputs
        hidden_list = []
        for l in range(self.num_layers):
            h = self.gat_layers[l](g, h)
            hidden_list.append(h)
            # h = h.flatten(1)
        # output projection
        if return_hidden:
            return self.head(h), hidden_list
        else:
            return self.head(h)

    def reset_classifier(self, num_classes):
        self.head = nn.Linear(self.num_heads * self.out_dim, num_classes)


class DotGatConv(nn.Module):
    def __init__(self,
                 in_feats,
                 out_feats,
                 num_heads,
                 feat_drop,
                 attn_drop,
                 residual,
                 activation=None,
                 norm=None,
                 concat_out=False,
                 allow_zero_in_degree=False):
        super(DotGatConv, self).__init__()
        self._in_src_feats, self._in_dst_feats = expand_as_pair(in_feats)
        self._out_feats = out_feats
        self._allow_zero_in_degree = allow_zero_in_degree
        self._num_heads = num_heads
        self._concat_out = concat_out

        self.feat_drop = nn.Dropout(feat_drop)
        self.attn_drop = nn.Dropout(attn_drop) if attn_drop > 0 else nn.Identity()
        self.activation = activation

        if isinstance(in_feats, tuple):
            self.fc_src = nn.Linear(self._in_src_feats, self._out_feats*self._num_heads, bias=False)
            self.fc_dst = nn.Linear(self._in_dst_feats, self._out_feats*self._num_heads, bias=False)
        else:
            self.fc = nn.Linear(self._in_src_feats, self._out_feats*self._num_heads, bias=False)

        if residual:
            if self._in_dst_feats != out_feats * num_heads:
                self.res_fc = nn.Linear(
                    self._in_dst_feats, num_heads * out_feats, bias=False)
            else:
                self.res_fc = nn.Identity()
        else:
            self.register_buffer('res_fc', None)

        self.norm = norm
        if norm is not None:
            self.norm = norm(num_heads * out_feats)

    def forward(self, graph, feat, get_attention=False):
        r"""

        Description
        -----------
        Apply dot product version of self attention in GCN.

        Parameters
        ----------
        graph: DGLGraph or bi_partities graph
            The graph
        feat: torch.Tensor or pair of torch.Tensor
            If a torch.Tensor is given, the input feature of shape :math:`(N, D_{in})` where
            :math:`D_{in}` is size of input feature, :math:`N` is the number of nodes.
            If a pair of torch.Tensor is given, the pair must contain two tensors of shape
            :math:`(N_{in}, D_{in_{src}})` and :math:`(N_{out}, D_{in_{dst}})`.
        get_attention : bool, optional
            Whether to return the attention values. Default to False.

        Returns
        -------
        torch.Tensor
            The output feature of shape :math:`(N, D_{out})` where :math:`D_{out}` is size
            of output feature.
        torch.Tensor, optional
            The attention values of shape :math:`(E, 1)`, where :math:`E` is the number of
            edges. This is returned only when :attr:`get_attention` is ``True``.

        Raises
        ------
        DGLError
            If there are 0-in-degree nodes in the input graph, it will raise DGLError
            since no message will be passed to those nodes. This will cause invalid output.
            The error can be ignored by setting ``allow_zero_in_degree`` parameter to ``True``.
        """

        graph = graph.local_var()

        if not self._allow_zero_in_degree:
            if (graph.in_degrees() == 0).any():
                raise ValueError('There are 0-in-degree nodes in the graph, '
                               'output for those nodes will be invalid. '
                               'This is harmful for some applications, '
                               'causing silent performance regression. '
                               'Adding self-loop on the input graph by '
                               'calling `g = dgl.add_self_loop(g)` will resolve '
                               'the issue. Setting ``allow_zero_in_degree`` '
                               'to be `True` when constructing this module will '
                               'suppress the check and let the code run.')

        # check if feat is a tuple
        if isinstance(feat, tuple):
            h_src = feat[0]
            h_dst = feat[1]
            feat_src = self.fc_src(h_src).view(-1, self._num_heads, self._out_feats)
            feat_dst = self.fc_dst(h_dst).view(-1, self._num_heads, self._out_feats)
            print("!! tuple input in DotGAT !!")
        else:
            feat = self.feat_drop(feat)
            h_src = feat
            feat_src = feat_dst = self.fc(h_src).view(-1, self._num_heads, self._out_feats)
            if graph.is_block:
                feat_dst = feat_src[:graph.number_of_dst_nodes()]

        # Assign features to nodes
        graph.srcdata.update({'ft': feat_src})
        graph.dstdata.update({'ft': feat_dst})

        # Step 1. dot product
        graph.apply_edges(fn.u_dot_v('ft', 'ft', 'a'))

        # Step 2. edge softmax to compute attention scores
        graph.edata['sa'] = edge_softmax(graph, graph.edata['a'] / self._out_feats**0.5)
        graph.edata["sa"] = self.attn_drop(graph.edata["sa"])
        # Step 3. Broadcast softmax value to each edge, and aggregate dst node
        graph.update_all(fn.u_mul_e('ft', 'sa', 'attn'), fn.sum('attn', 'agg_u'))

        # output results to the destination nodes
        rst = graph.dstdata['agg_u']

        if self.res_fc is not None:
            # Use -1 rather than self._num_heads to handle broadcasting
            batch_size = feat.shape[0]
            resval = self.res_fc(h_dst).view(batch_size, -1, self._out_feats)
            rst = rst + resval

        if self._concat_out:
            rst = rst.flatten(1)
        else:
            rst = torch.mean(rst, dim=1)

        if self.norm is not None:
            rst = self.norm(rst)

        # activation
        if self.activation:
            rst = self.activation(rst)

        if get_attention:
            return rst, graph.edata['sa']
        else:
            return rst



import copy
from tqdm import tqdm
import torch
import torch.nn as nn
import numpy as np
from utils import create_optimizer, accuracy

def node_classification_evaluation(model, graph, x, num_classes, lr_f, weight_decay_f, max_epoch_f, device, linear_prob=True, mute=False):
    model.eval()
    if linear_prob:
        with torch.no_grad():
            x = model.embed(graph.to(device), x.to(device))
            in_feat = x.shape[1]
        encoder = LogisticRegression(in_feat, num_classes)
    else:
        encoder = model.encoder
        encoder.reset_classifier(num_classes)

    num_finetune_params = [p.numel() for p in encoder.parameters() if  p.requires_grad]
    if not mute:
        print(f"num parameters for finetuning: {sum(num_finetune_params)}")
    
    encoder.to(device)
    optimizer_f = create_optimizer("adam", encoder, lr_f, weight_decay_f)
    #final_acc, estp_acc,best_model = linear_probing_for_transductive_node_classiifcation(encoder, graph, x, optimizer_f, max_epoch_f, device, mute)
    labels = graph.ndata["label"]
    mask = torch.tensor( np.arange(0,graph.num_nodes(),1))

    final_acc, estp_acc,best_model = linear_probing_for_transductive_node_classiifcation(encoder, graph, x, optimizer_f, max_epoch_f, device, mute)
    return final_acc, estp_acc, best_model


def node_classification_out_model_evaluation(model, graph, x, num_classes, lr_f, weight_decay_f, max_epoch_f, device, linear_prob=True, mute=False):
    model.eval()
    if linear_prob:
        with torch.no_grad():
            x = model.embed(graph.to(device), x.to(device))
            in_feat = x.shape[1]
        encoder = LogisticRegression(in_feat, num_classes)
    else:
        encoder = model.encoder
        encoder.reset_classifier(num_classes)

    num_finetune_params = [p.numel() for p in encoder.parameters() if  p.requires_grad]
    if not mute:
        print(f"num parameters for finetuning: {sum(num_finetune_params)}")
    
    encoder.to(device)
    optimizer_f = create_optimizer("adam", encoder, lr_f, weight_decay_f)
    #final_acc, estp_acc,best_model = linear_probing_for_transductive_node_classiifcation(encoder, graph, x, optimizer_f, max_epoch_f, device, mute)
    labels = graph.ndata["label"]
    mask = torch.tensor( np.arange(0,graph.num_nodes(),1))

    final_acc, estp_acc,best_model = linear_probing_for_inductive_node_classiifcation(encoder, x, labels, mask,optimizer_f, max_epoch_f, device, mute)
    return final_acc, estp_acc, best_model


def linear_probing_for_transductive_node_classiifcation(model, graph, feat, optimizer, max_epoch, device, mute=False):
    criterion = torch.nn.CrossEntropyLoss()

    graph = graph.to(device)
    x = feat.to(device)

    train_mask = torch.tensor( np.arange(0,graph.num_nodes(),1))
    val_mask = torch.tensor(np.arange(0,graph.num_nodes(),1))
    test_mask = torch.tensor(np.arange(0,graph.num_nodes(),1))
    labels = graph.ndata["label"]

    best_val_acc = 0
    best_val_epoch = 0
    best_model = None

    if not mute:
        epoch_iter = tqdm(range(max_epoch))
    else:
        epoch_iter = range(max_epoch)

    for epoch in epoch_iter:
        model.train()
        out = model(graph, x)
        loss = criterion(out[train_mask], labels[train_mask])
        optimizer.zero_grad()
        loss.backward()
        # torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=3)
        optimizer.step()

        with torch.no_grad():
            model.eval()
            pred = model(graph, x)
            val_acc = accuracy(pred[val_mask], labels[val_mask])
            val_loss = criterion(pred[val_mask], labels[val_mask])
            test_acc = accuracy(pred[test_mask], labels[test_mask])
            test_loss = criterion(pred[test_mask], labels[test_mask])
        
        if val_acc >= best_val_acc:
            best_val_acc = val_acc
            best_val_epoch = epoch
            best_model = copy.deepcopy(model)

        if not mute:
            epoch_iter.set_description(f"# Epoch: {epoch}, train_loss:{loss.item(): .4f}, val_loss:{val_loss.item(): .4f}, val_acc:{val_acc}, test_loss:{test_loss.item(): .4f}, test_acc:{test_acc: .4f}")

    best_model.eval()
    with torch.no_grad():
        pred = best_model(graph, x)
        estp_test_acc = accuracy(pred[test_mask], labels[test_mask])
    if mute:
        print(f"# IGNORE: --- TestAcc: {test_acc:.4f}, early-stopping-TestAcc: {estp_test_acc:.4f}, Best ValAcc: {best_val_acc:.4f} in epoch {best_val_epoch} --- ")
    else:
        print(f"--- TestAcc: {test_acc:.4f}, early-stopping-TestAcc: {estp_test_acc:.4f}, Best ValAcc: {best_val_acc:.4f} in epoch {best_val_epoch} --- ")

    # (final_acc, es_acc, best_acc)
    return test_acc, estp_test_acc,best_model


def linear_probing_for_inductive_node_classiifcation(model, x, labels, mask, optimizer, max_epoch, device, mute=False):
    if len(labels.shape) > 1:
        criterion = torch.nn.BCEWithLogitsLoss()
    else:
        criterion = torch.nn.CrossEntropyLoss()
    train_mask= mask
    val_mask = mask
    test_mask = mask

    best_val_acc = 0
    best_val_epoch = 0
    best_model = None

    if not mute:
        epoch_iter = tqdm(range(max_epoch))
    else:
        epoch_iter = range(max_epoch)  

        best_val_acc = 0

    if not mute:
        epoch_iter = tqdm(range(max_epoch))
    else:
        epoch_iter = range(max_epoch)

    for epoch in epoch_iter:
        model.train()
        out = model(None, x)
        loss = criterion(out[train_mask], labels[train_mask])
        optimizer.zero_grad()
        loss.backward()
        # torch.nn.utils.clip_grad_norm_(model.parameters(), max_norm=3)
        optimizer.step()

        with torch.no_grad():
            model.eval()
            pred = model(None, x)
            val_acc = accuracy(pred[val_mask], labels[val_mask])
            val_loss = criterion(pred[val_mask], labels[val_mask])
            test_acc = accuracy(pred[test_mask], labels[test_mask])
            test_loss = criterion(pred[test_mask], labels[test_mask])
        
        if val_acc >= best_val_acc:
            best_val_acc = val_acc
            best_val_epoch = epoch
            best_model = copy.deepcopy(model)

        if not mute:
            epoch_iter.set_description(f"# Epoch: {epoch}, train_loss:{loss.item(): .4f}, val_loss:{val_loss.item(): .4f}, val_acc:{val_acc}, test_loss:{test_loss.item(): .4f}, test_acc:{test_acc: .4f}")

    best_model.eval()
    with torch.no_grad():
        pred = best_model(None, x)
        estp_test_acc = accuracy(pred[test_mask], labels[test_mask])
    if mute:
        print(f"# IGNORE: --- TestAcc: {test_acc:.4f}, early-stopping-TestAcc: {estp_test_acc:.4f}, Best ValAcc: {best_val_acc:.4f} in epoch {best_val_epoch} ")
    else:
        print(f"--- TestAcc: {test_acc:.4f}, early-stopping-TestAcc: {estp_test_acc:.4f}, Best ValAcc: {best_val_acc:.4f} in epoch {best_val_epoch}")

    return test_acc, estp_test_acc,best_model



class LogisticRegression(nn.Module):
    def __init__(self, num_dim, num_class):
        super().__init__()
        self.linear = nn.Linear(num_dim, num_class)

    def forward(self, g, x, *args):
        logits = self.linear(x)
        return logits



from sklearn.metrics import (adjusted_rand_score, calinski_harabasz_score)
import logging
from tqdm import tqdm
from utils import (
    create_optimizer,
    set_random_seed,
    get_current_lr,
)
from graphmae.models import build_model
from sklearn.cluster import KMeans
import scanpy as sc
from operator import index
from scipy.spatial.distance import pdist, squareform
import pandas as pd
import numpy as np
from scipy import sparse
import dgl
import torch
import warnings
import argparse

warnings.filterwarnings("ignore")
class myDict(object):
    def __init__(self,mydict):
        self.mydict = mydict
        self.length = []
        self.keys = []
        for key,values in self.mydict.items():
            self.keys.append(key)
            self.length.append(len(values))
        self.nums = [1] * len(self.length)
        for i in range(len(self.length)):
            for j in range(i,len(self.length)):
                self.nums[i] *= self.length[j]
        self.para_dis = []
        print(self.length)
        print(self.nums)
                
    def getindex(self,index):
        result = []
        value = index
        for i in range(len(self.nums) - 1):
            result.append(value // self.nums[i+1])
            value = value - result[i] * self.nums[i+1]
        result.append(value) 
        result_dict = dict()
        for index,value in enumerate(result):
            result_dict[self.keys[index]] = self.mydict.get(self.keys[index])[value]
        return result_dict
    
    #para_dis = []
    def myiter(self):
        #para_dis = []
        for i in range(0,self.nums[0]):
            self.para_dis.append(self.getindex(i))
        return self.para_dis
def kMeans_use(embedding,cluster_number):
    kmeans = KMeans(n_clusters=cluster_number,
                init="k-means++",
                random_state=0)
    pred = kmeans.fit_predict(embedding)
    return pred
def pretrain(model, graph, feat, optimizer, max_epoch, device, scheduler, logger=None):
    logging.info("start training..")
    graph = graph.to(device)
    x = feat.to(device)

    epoch_iter = tqdm(range(max_epoch))

    for epoch in epoch_iter:
        model.train()

        loss, loss_dict = model(graph, x)

        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        if scheduler is not None:
            scheduler.step()

        epoch_iter.set_description(f"# Epoch {epoch}: train_loss: {loss.item():.4f}")
        if logger is not None:
            loss_dict["lr"] = get_current_lr(optimizer)
            logger.note(loss_dict, step=epoch)
    return model
def drawPicture(dataframe,col_name, row_name,colorattribute,save_file,celltype_colors =  ("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3", "#FF7F00", "#FFFF33", "#A65628", "#F781BF", "#999999", "#E41A80",
    "#377F1C", "#4DAFAE", "#984F07", "#FF7F64", "#FFFF97", "#A6568C", "#F78223", "#9999FD", "#E41AE4", "#377F80", "#4DB012",
    "#984F6B", "#FF7FC8",  "#A656F0", "#F78287", "#999A61", "#E41B48", "#377FE4", "#4DB076", "#984FCF", "#FF802C",
    "#00005F", "#A65754", "#F782EB", "#999AC5", "#E41BAC", "#378048", "#4DB0DA", "#985033", "#FF8090", "#0000C3", "#A657B8",
    "#F7834F", "#999B29", "#E41C10", "#3780AC", "#4DB13E", "#985097", "#FF80F4", "#000127", "#A6581C", "#F783B3", "#999B8D",
    "#E41C74", "#378110", "#4DB1A2", "#9850FB", "#FF8158", "#00018B", "#A65880", "#F78417", "#999BF1", "#E41CD8", "#378174",
    "#4DB206", "#98515F", "#FF81BC", "#0001EF", "#A658E4", "#F7847B", "#999C55", "#E41D3C", "#3781D8", "#4DB26A", "#9851C3",
    "#FF8220", "#000253", "#A65948", "#F784DF", "#999CB9", "#E41DA0", "#37823C", "#4DB2CE", "#985227", "#FF8284", "#0002B7",
    "#A659AC", "#F78543", "#999D1D", "#E41E04", "#3782A0", "#4DB332", "#98528B", "#FF82E8", "#00031B", "#A65A10", "#F785A7",
    "#999D81", "#E41E68", "#378304", "#4DB396", "#9852EF", "#FF834C", "#00037F", "#A65A74", "#F7860B", "#999DE5", "#E41ECC",
    "#378368", "#4DB3FA", "#985353", "#FF83B0", "#0003E3", "#A65AD8", "#F7866F", "#999E49", "#E41F30", "#3783CC", "#4DB45E",
    "#9853B7", "#FF8414", "#000447", "#A65B3C", "#F786D3", "#999EAD", "#E41F94", "#378430", "#4DB4C2", "#98541B", "#FF8478",
    "#0004AB", "#A65BA0", "#F78737", "#999F11", "#E41FF8", "#378494", "#4DB526", "#98547F", "#FF84DC", "#00050F", "#A65C04",
    "#F7879B", "#999F75"
    ),width = 1000,height = 1000,marker_size = 14,is_show = True,is_save = False,save_type = "pdf"):

    import plotly.express as px
    size = len(set(dataframe[colorattribute]))
    #dataframe.sort_values(by = colorattribute,inplace=True, ascending=True)
    length_col = max(dataframe[col_name]) - min(dataframe[col_name])

    length_row = max(dataframe[row_name]) - min(dataframe[row_name])
    max_length = max(length_col,length_row) + 2
    fig = px.scatter(dataframe, x = col_name, y= row_name,color = colorattribute,color_discrete_sequence=celltype_colors)
    fig.update_traces(marker_size=marker_size)
    fig.update_layout(
        xaxis = dict(
            tickmode = 'linear',
            tick0 = min(dataframe[row_name]),   # 起始点
            dtick = max_length  # 间距
        ),
        yaxis = dict(
            tickmode = 'linear',
            tick0 = min(dataframe[col_name]),   # 起始点
            dtick = max_length  # 间距
        ),
    #     xaxis_range = [min(it_mapping_csv.row)-10,min(it_mapping_csv.row) + max_length ],
    #     yaxis_range =[min(it_mapping_csv.col)-10,min(it_mapping_csv.col) + max_length]
    )
    fig.update_layout(
        autosize=False,
        width=width,
        height = height)
    if(is_show):
        fig.show()
    if(is_save):
        if(save_type == "pdf"):
            fig.write_image(save_file)
        if(save_type == "html"):
            fig.write_html(save_file)
def mclust_R(adata, num_cluster,obs_name = "mclust" ,modelNames='EEE', used_obsm='scGraphMAE',random_seed = 2022):
    """\
    Clustering using the mclust algorithm.
    The parameters are the same as those in the R package mclust.
    """
    
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster,modelNames)
    #print(mclust_res)
    mclust_res = np.array(res[-2])
    #print(mclust_res)
    adata.obs[obs_name] = mclust_res
    adata.obs[obs_name] = adata.obs[obs_name].astype('int')
    adata.obs[obs_name] = adata.obs[obs_name].astype('category')
    return adata
def Graph_get(adata,threshold = 30,feature_dim_method = "PCA",num_features = 600,row_name = "imagerow",col_name = "imagecol"):
    cell_loc = adata.obsm["spatial"]
    distance_np = pdist(cell_loc, metric = "euclidean")
    distance_np_X =squareform(distance_np)
    #distance_loc_csv = pd.DataFrame(index=adata.obs.index, columns=adata.obs.index,data = distance_np_X)
    num_big = np.where((0< distance_np_X)&(distance_np_X < threshold))[0].shape[0]
    adj_matrix = np.zeros(distance_np_X.shape)
    non_zero_point = np.where((0 < distance_np_X) & (distance_np_X < threshold))
    for i in range(num_big):
        x = non_zero_point[0][i]
        y = non_zero_point[1][i]
        adj_matrix[x][y] = 1 
    adj_matrix = adj_matrix + np.eye(distance_np_X.shape[0])
    adj_matrix  = np.float32(adj_matrix)
    adj_matrix_crs = sparse.csr_matrix(adj_matrix)
    graph = dgl.from_scipy(adj_matrix_crs,eweight_name='w')
    if(feature_dim_method == "PCA"):
        adata_X = adata.obsm["X_pca"]
    else:
        sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=num_features)
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        adata_Vars =  adata[:, adata.var['highly_variable']]
        adata_X = adata_Vars.X.todense()
    graph.ndata["feat"] = torch.tensor(adata_X.copy())
    print(graph)
    return graph
def train_model(adata,graph,args,device = 0 ,seed = 0):
    max_epoch = args.max_epoch
    #训练模型
    #提取到的embedding保存在adata里的obsm["scGraphMAE"]、
    lr = args.lr
    weight_decay = args.weight_decay
    optim_type = args.optimizer 
    set_random_seed(seed)
    model = build_model(args)
    model.to(device)
    optimizer = create_optimizer(optim_type, model, lr, weight_decay)
    x = graph.ndata["feat"]
    scheduler  = None
    logger = None
    if not load_model:
        model = pretrain(model, graph, x, optimizer, max_epoch, device, scheduler)
    embedding = model.embed(graph.to(device), x.to(device))
    adata.obsm["scGraphMAE"] = embedding.cpu().detach().numpy()
    return adata
def cluster_comp(adata,num_classes, used_obsm='scGraphMAE',K_means_value = "K_means_value",m_clust_EEE = "mclust_EEE",m_clust_VII = "mclust_VII",m_clust_EII = "mclust_EII"):
    k_means_value = kMeans_use(adata.obsm[used_obsm],num_classes)
    adata.obs[K_means_value] = k_means_value
    #k_means_score = adjusted_rand_score(adata.obs.lay_num.values, k_means_value )
    try:
        adata = mclust_R(adata,used_obsm=used_obsm,obs_name = m_clust_EEE,num_cluster=num_classes,modelNames = "EEE")
    except:
        adata.obs[m_clust_EEE] = 0
    try:
        adata = mclust_R(adata,used_obsm=used_obsm,obs_name = m_clust_VII,num_cluster=num_classes,modelNames = "VII")
    except:
        adata.obs[m_clust_VII] = 0
    try:
        adata = mclust_R(adata,used_obsm=used_obsm,obs_name = m_clust_EII,num_cluster=num_classes,modelNames = "EII")
    except:
        adata.obs[m_clust_EII] = 0
    return adata
def evaluate_ari_comp(adata,true_columns = "lay_num",K_means_value = "K_means_value",m_clust_EEE = "mclust_EEE",m_clust_VII = "mclust_VII",m_clust_EII = "mclust_EII"):
    if(K_means_value != None):
        k_means_score = adjusted_rand_score(adata.obs[true_columns], adata.obs[K_means_value] )
        adata.uns[K_means_value] = k_means_score
    if(m_clust_EEE != None):
        m_clust_EEE_score = adjusted_rand_score(adata.obs[true_columns], adata.obs[m_clust_EEE] )
        adata.uns[m_clust_EEE] = m_clust_EEE_score
    if(m_clust_VII != None):
        m_clust_VII_score = adjusted_rand_score(adata.obs[true_columns], adata.obs[m_clust_VII] )
        adata.uns[m_clust_VII] = m_clust_VII_score
    if(m_clust_EII != None):
        m_clust_EII_score = adjusted_rand_score(adata.obs[true_columns], adata.obs[m_clust_EII] )
        adata.uns[m_clust_EII] = m_clust_EII_score
    return adata
parser = argparse.ArgumentParser(description="GAT")
parser.add_argument("--seeds", type=int, nargs="+", default=[0])
parser.add_argument("--dataset", type=str, default="cora")
parser.add_argument("--device", type=int, default=-1)
parser.add_argument("--max_epoch", type=int, default=200,
                    help="number of training epochs")
parser.add_argument("--warmup_steps", type=int, default=-1)

parser.add_argument("--num_heads", type=int, default=4,
                    help="number of hidden attention heads")
parser.add_argument("--num_out_heads", type=int, default=1,
                    help="number of output attention heads")
parser.add_argument("--num_layers", type=int, default=2,
                    help="number of hidden layers")
parser.add_argument("--num_hidden", type=int, default=256,
                    help="number of hidden units")
parser.add_argument("--residual", action="store_true", default=False,
                    help="use residual connection")
parser.add_argument("--in_drop", type=float, default=.2,
                    help="input feature dropout")
parser.add_argument("--attn_drop", type=float, default=.1,
                    help="attention dropout")
parser.add_argument("--norm", type=str, default=None)
parser.add_argument("--lr", type=float, default=0.005,
                    help="learning rate")
parser.add_argument("--weight_decay", type=float, default=5e-4,
                    help="weight decay")
parser.add_argument("--negative_slope", type=float, default=0.2,
                    help="the negative slope of leaky relu for GAT")
parser.add_argument("--activation", type=str, default="prelu")
parser.add_argument("--mask_rate", type=float, default=0.5)
parser.add_argument("--drop_edge_rate", type=float, default=0.0)
parser.add_argument("--replace_rate", type=float, default=0.0)

parser.add_argument("--encoder", type=str, default="gat")
parser.add_argument("--decoder", type=str, default="gat")
parser.add_argument("--loss_fn", type=str, default="byol")
parser.add_argument("--alpha_l", type=float, default=2, help="`pow`inddex for `sce` loss")
parser.add_argument("--optimizer", type=str, default="adam")

parser.add_argument("--max_epoch_f", type=int, default=30)
parser.add_argument("--lr_f", type=float, default=0.001, help="learning rate for evaluation")
parser.add_argument("--weight_decay_f", type=float, default=0.0, help="weight decay for evaluation")
parser.add_argument("--linear_prob", action="store_true", default=False)

parser.add_argument("--load_model", action="store_true")
parser.add_argument("--save_model", action="store_true")
parser.add_argument("--use_cfg", action="store_true")
parser.add_argument("--logging", action="store_true")
parser.add_argument("--scheduler", action="store_true", default=False)
parser.add_argument("--concat_hidden", action="store_true", default=False)

# for graph classification
parser.add_argument("--pooling", type=str, default="mean")
parser.add_argument("--deg4feat", action="store_true", default=False, help="use node degree as input feature")
parser.add_argument("--batch_size", type=int, default=32)
sample_names = ["151507","151508","151509","151510","151669","151670","151671","151672","151673","151674","151675","151676"]
#sample_names = ["151507"]
scGraph_fold = "/home/sunhang/Embedding/baseline/10X_batch/"
result_csv = pd.DataFrame()


import torch
import torch.nn as nn
import torch.nn.functional as F

import dgl.function as fn
from dgl.utils import expand_as_pair

from tools.utils import create_activation


class GCN(nn.Module):
    def __init__(self,
                 in_dim,
                 num_hidden,
                 out_dim,
                 num_layers,
                 dropout,
                 activation,
                 residual,
                 norm,
                 encoding=False
                 ):
        super(GCN, self).__init__()
        self.out_dim = out_dim
        self.num_layers = num_layers
        self.gcn_layers = nn.ModuleList()
        self.activation = activation
        self.dropout = dropout

        last_activation = create_activation(activation) if encoding else None
        last_residual = encoding and residual
        last_norm = norm if encoding else None
        
        if num_layers == 1:
            self.gcn_layers.append(GraphConv(
                in_dim, out_dim, residual=last_residual, norm=last_norm, activation=last_activation))
        else:
            # input projection (no residual)
            self.gcn_layers.append(GraphConv(
                in_dim, num_hidden, residual=residual, norm=norm, activation=create_activation(activation)))
            # hidden layers
            for l in range(1, num_layers - 1):
                # due to multi-head, the in_dim = num_hidden * num_heads
                self.gcn_layers.append(GraphConv(
                    num_hidden, num_hidden, residual=residual, norm=norm, activation=create_activation(activation)))
            # output projection
            self.gcn_layers.append(GraphConv(
                num_hidden, out_dim, residual=last_residual, activation=last_activation, norm=last_norm))

        # if norm is not None:
        #     self.norms = nn.ModuleList([
        #         norm(num_hidden)
        #         for _ in range(num_layers - 1)
        #     ])
        #     if not encoding:
        #         self.norms.append(norm(out_dim))
        # else:
        #     self.norms = None
        self.norms = None
        self.head = nn.Identity()

    def forward(self, g, inputs, return_hidden=False):
        h = inputs
        hidden_list = []
        for l in range(self.num_layers):
            h = F.dropout(h, p=self.dropout, training=self.training)
            h = self.gcn_layers[l](g, h)
            if self.norms is not None and l != self.num_layers - 1:
                h = self.norms[l](h)
            hidden_list.append(h)
        # output projection
        if self.norms is not None and len(self.norms) == self.num_layers:
            h = self.norms[-1](h)
        if return_hidden:
            return self.head(h), hidden_list
        else:
            return self.head(h)

    def reset_classifier(self, num_classes):
        self.head = nn.Linear(self.out_dim, num_classes)


class GraphConv(nn.Module):
    def __init__(self,
                 in_dim,
                 out_dim,
                 norm=None,
                 activation=None,
                 residual=True,
                 ):
        super().__init__()
        self._in_feats = in_dim
        self._out_feats = out_dim

        self.fc = nn.Linear(in_dim, out_dim)

        if residual:
            if self._in_feats != self._out_feats:
                self.res_fc = nn.Linear(
                    self._in_feats, self._out_feats, bias=False)
                print("! Linear Residual !")
            else:
                print("Identity Residual ")
                self.res_fc = nn.Identity()
        else:
            self.register_buffer('res_fc', None)

        # if norm == "batchnorm":
        #     self.norm = nn.BatchNorm1d(out_dim)
        # elif norm == "layernorm":
        #     self.norm = nn.LayerNorm(out_dim)
        # else:
        #     self.norm = None
        
        self.norm = norm
        if norm is not None:
            self.norm = norm(out_dim)
        self._activation = activation

        self.reset_parameters()

    def reset_parameters(self):
        self.fc.reset_parameters()

    def forward(self, graph, feat):
        with graph.local_scope():
            aggregate_fn = fn.copy_src('h', 'm')
            # if edge_weight is not None:
            #     assert edge_weight.shape[0] == graph.number_of_edges()
            #     graph.edata['_edge_weight'] = edge_weight
            #     aggregate_fn = fn.u_mul_e('h', '_edge_weight', 'm')

            # (BarclayII) For RGCN on heterogeneous graphs we need to support GCN on bipartite.
            feat_src, feat_dst = expand_as_pair(feat, graph)
            # if self._norm in ['left', 'both']:
            degs = graph.out_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5)
            shp = norm.shape + (1,) * (feat_src.dim() - 1)
            norm = torch.reshape(norm, shp)
            feat_src = feat_src * norm

            # if self._in_feats > self._out_feats:
            #     # mult W first to reduce the feature size for aggregation.
            #     # if weight is not None:
            #         # feat_src = th.matmul(feat_src, weight)
            #     graph.srcdata['h'] = feat_src
            #     graph.update_all(aggregate_fn, fn.sum(msg='m', out='h'))
            #     rst = graph.dstdata['h']
            # else:
            # aggregate first then mult W
            graph.srcdata['h'] = feat_src
            graph.update_all(aggregate_fn, fn.sum(msg='m', out='h'))
            rst = graph.dstdata['h']
            
            rst = self.fc(rst)

            # if self._norm in ['right', 'both']:
            degs = graph.in_degrees().float().clamp(min=1)
            norm = torch.pow(degs, -0.5)
            shp = norm.shape + (1,) * (feat_dst.dim() - 1)
            norm = torch.reshape(norm, shp)
            rst = rst * norm

            if self.res_fc is not None:
                rst = rst + self.res_fc(feat_dst)

            if self.norm is not None:
                rst = self.norm(rst)

            if self._activation is not None:
                rst = self._activation(rst)

            return rst


from .edcoder import PreModel


def build_model(args):
    num_heads = args.num_heads
    num_out_heads = args.num_out_heads
    num_hidden = args.num_hidden
    num_layers = args.num_layers
    residual = args.residual
    attn_drop = args.attn_drop
    in_drop = args.in_drop
    norm = args.norm
    negative_slope = args.negative_slope
    encoder_type = args.encoder
    decoder_type = args.decoder
    mask_rate = args.mask_rate
    drop_edge_rate = args.drop_edge_rate
    replace_rate = args.replace_rate


    activation = args.activation
    loss_fn = args.loss_fn
    alpha_l = args.alpha_l
    concat_hidden = args.concat_hidden
    num_features = args.num_features


    model = PreModel(
        in_dim=num_features,
        num_hidden=num_hidden,
        num_layers=num_layers,
        nhead=num_heads,
        nhead_out=num_out_heads,
        activation=activation,
        feat_drop=in_drop,
        attn_drop=attn_drop,
        negative_slope=negative_slope,
        residual=residual,
        encoder_type=encoder_type,
        decoder_type=decoder_type,
        mask_rate=mask_rate,
        norm=norm,
        loss_fn=loss_fn,
        drop_edge_rate=drop_edge_rate,
        replace_rate=replace_rate,
        alpha_l=alpha_l,
        concat_hidden=concat_hidden,
    )
    return model




from collections import namedtuple, Counter
import numpy as np

import torch
import torch.nn.functional as F

import dgl
from dgl.data import (
    load_data, 
    TUDataset, 
    CoraGraphDataset, 
    CiteseerGraphDataset, 
    PubmedGraphDataset
)
from ogb.nodeproppred import DglNodePropPredDataset
from dgl.data.ppi import PPIDataset
from dgl.dataloading import GraphDataLoader

from sklearn.preprocessing import StandardScaler


GRAPH_DICT = {
    "cora": CoraGraphDataset,
    "citeseer": CiteseerGraphDataset,
    "pubmed": PubmedGraphDataset,
    "ogbn-arxiv": DglNodePropPredDataset
}


def preprocess(graph):
    feat = graph.ndata["feat"]
    graph = dgl.to_bidirected(graph)
    graph.ndata["feat"] = feat

    graph = graph.remove_self_loop().add_self_loop()
    graph.create_formats_()
    return graph


def scale_feats(x):
    scaler = StandardScaler()
    feats = x.numpy()
    scaler.fit(feats)
    feats = torch.from_numpy(scaler.transform(feats)).float()
    return feats


def load_dataset(dataset_name):
    assert dataset_name in GRAPH_DICT, f"Unknow dataset: {dataset_name}."
    if dataset_name.startswith("ogbn"):
        dataset = GRAPH_DICT[dataset_name](dataset_name)
    else:
        dataset = GRAPH_DICT[dataset_name]()

    if dataset_name == "ogbn-arxiv":
        graph, labels = dataset[0]
        num_nodes = graph.num_nodes()

        split_idx = dataset.get_idx_split()
        train_idx, val_idx, test_idx = split_idx["train"], split_idx["valid"], split_idx["test"]
        graph = preprocess(graph)

        if not torch.is_tensor(train_idx):
            train_idx = torch.as_tensor(train_idx)
            val_idx = torch.as_tensor(val_idx)
            test_idx = torch.as_tensor(test_idx)

        feat = graph.ndata["feat"]
        feat = scale_feats(feat)
        graph.ndata["feat"] = feat

        train_mask = torch.full((num_nodes,), False).index_fill_(0, train_idx, True)
        val_mask = torch.full((num_nodes,), False).index_fill_(0, val_idx, True)
        test_mask = torch.full((num_nodes,), False).index_fill_(0, test_idx, True)
        graph.ndata["label"] = labels.view(-1)
        graph.ndata["train_mask"], graph.ndata["val_mask"], graph.ndata["test_mask"] = train_mask, val_mask, test_mask
    else:
        graph = dataset[0]
        graph = graph.remove_self_loop()
        graph = graph.add_self_loop()
    num_features = graph.ndata["feat"].shape[1]
    num_classes = dataset.num_classes
    return graph, (num_features, num_classes)


def load_inductive_dataset(dataset_name):
    if dataset_name == "ppi":
        batch_size = 2
        # define loss function
        # create the dataset
        train_dataset = PPIDataset(mode='train')
        valid_dataset = PPIDataset(mode='valid')
        test_dataset = PPIDataset(mode='test')
        train_dataloader = GraphDataLoader(train_dataset, batch_size=batch_size)
        valid_dataloader = GraphDataLoader(valid_dataset, batch_size=batch_size, shuffle=False)
        test_dataloader = GraphDataLoader(test_dataset, batch_size=batch_size, shuffle=False)
        eval_train_dataloader = GraphDataLoader(train_dataset, batch_size=batch_size, shuffle=False)
        g = train_dataset[0]
        num_classes = train_dataset.num_labels
        num_features = g.ndata['feat'].shape[1]
    else:
        _args = namedtuple("dt", "dataset")
        dt = _args(dataset_name)
        batch_size = 1
        dataset = load_data(dt)
        num_classes = dataset.num_classes

        g = dataset[0]
        num_features = g.ndata["feat"].shape[1]

        train_mask = g.ndata['train_mask']
        feat = g.ndata["feat"]
        feat = scale_feats(feat)
        g.ndata["feat"] = feat

        g = g.remove_self_loop()
        g = g.add_self_loop()

        train_nid = np.nonzero(train_mask.data.numpy())[0].astype(np.int64)
        train_g = dgl.node_subgraph(g, train_nid)
        train_dataloader = [train_g]
        valid_dataloader = [g]
        test_dataloader = valid_dataloader
        eval_train_dataloader = [train_g]
        
    return train_dataloader, valid_dataloader, test_dataloader, eval_train_dataloader, num_features, num_classes



def load_graph_classification_dataset(dataset_name, deg4feat=False):
    dataset_name = dataset_name.upper()
    dataset = TUDataset(dataset_name)
    graph, _ = dataset[0]

    if "attr" not in graph.ndata:
        if "node_labels" in graph.ndata and not deg4feat:
            print("Use node label as node features")
            feature_dim = 0
            for g, _ in dataset:
                feature_dim = max(feature_dim, g.ndata["node_labels"].max().item())
            
            feature_dim += 1
            for g, l in dataset:
                node_label = g.ndata["node_labels"].view(-1)
                feat = F.one_hot(node_label, num_classes=feature_dim).float()
                g.ndata["attr"] = feat
        else:
            print("Using degree as node features")
            feature_dim = 0
            degrees = []
            for g, _ in dataset:
                feature_dim = max(feature_dim, g.in_degrees().max().item())
                degrees.extend(g.in_degrees().tolist())
            MAX_DEGREES = 400

            oversize = 0
            for d, n in Counter(degrees).items():
                if d > MAX_DEGREES:
                    oversize += n
            # print(f"N > {MAX_DEGREES}, #NUM: {oversize}, ratio: {oversize/sum(degrees):.8f}")
            feature_dim = min(feature_dim, MAX_DEGREES)

            feature_dim += 1
            for g, l in dataset:
                degrees = g.in_degrees()
                degrees[degrees > MAX_DEGREES] = MAX_DEGREES
                
                feat = F.one_hot(degrees, num_classes=feature_dim).float()
                g.ndata["attr"] = feat
    else:
        print("******** Use `attr` as node features ********")
        feature_dim = graph.ndata["attr"].shape[1]

    labels = torch.tensor([x[1] for x in dataset])
    
    num_classes = torch.max(labels).item() + 1
    dataset = [(g.remove_self_loop().add_self_loop(), y) for g, y in dataset]

    print(f"******** # Num Graphs: {len(dataset)}, # Num Feat: {feature_dim}, # Num Classes: {num_classes} ********")

    return dataset, (feature_dim, num_classes)



from sklearn.metrics import (adjusted_rand_score, normalized_mutual_info_score, 
                             silhouette_score, calinski_harabasz_score,
                             davies_bouldin_score,fowlkes_mallows_score)
from tqdm import tqdm
from models import build_model
import anndata as ad
import scanpy as sc
import pandas as pd
import numpy as np
import dgl
import torch
import os
import warnings
import matplotlib.pyplot as plt
from tools.utils import (
    create_optimizer,
    set_random_seed,
    pretrain,
    KMeans_use,
    drawPicture,
    mkdir,
)
warnings.filterwarnings("ignore")
from scipy.spatial.distance import pdist, squareform
from scipy import sparse
import argparse

def build_args():
    parser = argparse.ArgumentParser(description="GAT")
    parser.add_argument("--seeds", type=int, nargs="+", default=[0])
    parser.add_argument("--device", type=int, default=-1)
    parser.add_argument("--warmup_steps", type=int, default=-1)
    parser.add_argument("--num_heads", type=int, default=4,
                        help="number of hidden attention heads")
    parser.add_argument("--num_out_heads", type=int, default=1,
                        help="number of output attention heads")
    parser.add_argument("--residual", action="store_true", default=False,
                        help="use residual connection")
    parser.add_argument("--in_drop", type=float, default=0.2,
                        help="input feature dropout")
    parser.add_argument("--attn_drop", type=float, default=0.1,
                        help="attention dropout")
    parser.add_argument("--weight_decay", type=float, default=2e-4,
                        help="weight decay")
    parser.add_argument("--negative_slope", type=float, default=0.2,
                        help="the negative slope of leaky relu for GAT")
    parser.add_argument("--drop_edge_rate", type=float, default=0.0)
    parser.add_argument("--optimizer", type=str, default="adam")
    parser.add_argument("--max_epoch_f", type=int, default=300)
    parser.add_argument("--lr_f", type=float, default=0.01, help="learning rate for evaluation")
    parser.add_argument("--weight_decay_f", type=float, default=1e-4, help="weight decay for evaluation")
    parser.add_argument("--linear_prob", action="store_true", default=True)
    parser.add_argument("--load_model", action="store_true")
    parser.add_argument("--save_model", action="store_true")
    parser.add_argument("--use_cfg", action="store_true")
    parser.add_argument("--logging", action="store_true")
    parser.add_argument("--scheduler", action="store_true", default=True)
    parser.add_argument("--concat_hidden", action="store_true", default=False)
    # for graph classification
    parser.add_argument("--pooling", type=str, default="mean")
    parser.add_argument("--deg4feat", action="store_true", default=False, help="use node degree as input feature")
    parser.add_argument("--batch_size", type=int, default=32)
    # adjustable parameters
    parser.add_argument("--mask_rate", type=float, default=0.8)
    parser.add_argument("--encoder", type=str, default="gin")
    parser.add_argument("--decoder", type=str, default="gin")
    parser.add_argument("--num_hidden", type=int, default=128,
                        help="number of hidden units")
    parser.add_argument("--num_layers", type=int, default=3,
                        help="number of hidden layers")
    parser.add_argument("--activation", type=str, default="elu")
    parser.add_argument("--max_epoch", type=int, default=500,
                        help="number of training epochs")
    parser.add_argument("--lr", type=float, default=0.001 ,
                        help="learning rate")
    parser.add_argument("--alpha_l", type=float, default=4, help="`pow`inddex for `sce` loss")
    parser.add_argument("--loss_fn", type=str, default="sce")
    parser.add_argument("--replace_rate", type=float, default=0.05)
    parser.add_argument("--norm", type=str, default="batchnorm")
    # GSG parameter
    parser.add_argument("--feature_dim_method", type=str, default="PCA")
    parser.add_argument("--num_features", type=int, default=600)
    parser.add_argument("--threshold_radius", type=int, default=25)
    parser.add_argument("--folder_name", type=str, default="data/10X/")
    parser.add_argument("--sample_name", type=str, default="151673")
    parser.add_argument("--cluster_label", type=str, default= "")
    parser.add_argument("--num_classes", type=int, default=7,help = "The number of clusters")
    # read parameters
    args = parser.parse_args(args=[])
    return args

def GSG_Spatial_Pic(adata,args, result_file_path, is_show = False, is_save = True, spatial_figname = "/GSG_Cluster_Spatial.pdf", ground_truth_figname = "/Ground_true.pdf"):
    if(args.cluster_label != ""):
        adata.obs[args.cluster_label] = adata.obs[args.cluster_label].astype('category').cat.add_categories(['None'])
        adata.obs[args.cluster_label] = adata.obs[args.cluster_label].fillna("None")
        drawPicture(adata.obs,col_name ="imagecol",row_name = "imagerow",colorattribute=args.cluster_label,save_file = result_file_path + ground_truth_figname,is_show=is_show,is_save= is_save)
        drawPicture(adata.obs,col_name ="imagecol",row_name = "imagerow",colorattribute="GSG_Kmeans_cluster_str",save_file = result_file_path + spatial_figname,is_show=is_show,is_save= is_save)
        k_means_score_ari = adjusted_rand_score(adata.obs["GSG_Kmeans_cluster_str"].values, adata.obs[args.cluster_label].values)
        k_means_score_silhouette = silhouette_score(adata.obsm["GSG_embedding"], adata.obs["GSG_Kmeans_cluster_str"].values, metric="sqeuclidean")
        k_means_score_nmi = normalized_mutual_info_score(adata.obs["GSG_Kmeans_cluster_str"].values, adata.obs[args.cluster_label].values)
        k_means_score_fmi = fowlkes_mallows_score(adata.obs["GSG_Kmeans_cluster_str"].values, adata.obs[args.cluster_label].values)
        k_means_score_DB = davies_bouldin_score(adata.obsm["GSG_embedding"], adata.obs["GSG_Kmeans_cluster_str"].values)
        k_means_score_chs = calinski_harabasz_score(adata.obsm["GSG_embedding"], adata.obs["GSG_Kmeans_cluster_str"].values)
        print("k_means_score_ari:" + str(k_means_score_ari))
        print("k_means_score_silhouette:" + str(k_means_score_silhouette))
        print("k_means_score_nmi:" + str(k_means_score_nmi))
        print("k_means_score_fmi:" + str(k_means_score_fmi))
        print("k_means_score_DB:" + str(k_means_score_DB))
        print("k_means_score_chs:" + str(k_means_score_chs))
    else:
        drawPicture(adata.obs,col_name ="imagecol",row_name = "imagerow",colorattribute="GSG_Kmeans_cluster_str",save_file = result_file_path + spatial_figname,is_show=is_show,is_save= is_save)
        k_means_score_silhouette = silhouette_score(adata.obsm["GSG_embedding"], adata.obs["GSG_Kmeans_cluster_str"].values, metric="sqeuclidean")
        k_means_score_DB = davies_bouldin_score(adata.obsm["GSG_embedding"], adata.obs["GSG_Kmeans_cluster_str"].values)
        k_means_score_chs = calinski_harabasz_score(adata.obsm["GSG_embedding"], adata.obs["GSG_Kmeans_cluster_str"].values)
        print("k_means_score_silhouette:" + str(k_means_score_silhouette))
        print("k_means_score_DB:" + str(k_means_score_DB))
        print("k_means_score_chs:" + str(k_means_score_chs))
    return adata

def read_10X_Visium(path,
                    genome=None,
                    count_file='filtered_feature_bc_matrix.h5', 
                    library_id=None, 
                    load_images=True, 
                    quality='hires',
                    image_path = None):
    adata = sc.read_visium(path, 
                        genome=genome,
                        count_file=count_file,
                        library_id=library_id,
                        load_images=load_images,)
    adata.var_names_make_unique()
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]
    if quality == "fulres":
        image_coor = adata.obsm["spatial"]
        img = plt.imread(image_path, 0)
        adata.uns["spatial"][library_id]["images"]["fulres"] = img
    else:
        scale = adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + quality + "_scalef"]
        image_coor = adata.obsm["spatial"] * scale
    adata.obs["imagecol"] = image_coor[:, 0]
    adata.obs["imagerow"] = image_coor[:, 1]
    adata.uns["spatial"][library_id]["use_quality"] = quality
    return adata

def read_10X_Visium_with_label(path, 
                    genome=None,
                    count_file='filtered_feature_bc_matrix.h5', 
                    library_id=None, 
                    load_images=True, 
                    quality='hires',
                    image_path = None):
    adata = sc.read_visium(path, 
                        genome=genome,
                        count_file=count_file,
                        library_id=library_id,
                        load_images=load_images,)
    adata.var_names_make_unique()
    if library_id is None:
        library_id = list(adata.uns["spatial"].keys())[0]
    if quality == "fulres":
        image_coor = adata.obsm["spatial"]
        img = plt.imread(image_path, 0)
        adata.uns["spatial"][library_id]["images"]["fulres"] = img
    else:
        scale = adata.uns["spatial"][library_id]["scalefactors"][
            "tissue_" + quality + "_scalef"]
        image_coor = adata.obsm["spatial"] * scale
    if(os.path.exists(path + "/metadata.tsv")):
        adata.obs = pd.read_table(path + "/metadata.tsv",sep="\t",index_col=0)
    adata.obs["imagecol"] = image_coor[:, 0]
    adata.obs["imagerow"] = image_coor[:, 1]
    adata.uns["spatial"][library_id]["use_quality"] = quality
    return adata


def read_stereo_seq(counts_data_path, position_path):
    counts_file = os.path.join(counts_data_path)
    coor_file = os.path.join(position_path)
    coor_df = pd.read_csv(coor_file, sep='\t')
    counts = pd.read_csv(counts_file, sep='\t', index_col=0)
    counts.columns = ['Spot_' + str(x) for x in counts.columns]
    coor_df.index = coor_df['label'].map(lambda x: 'Spot_' + str(x))
    adata = sc.AnnData(counts.T)
    adata.obs = coor_df
    adata.var_names_make_unique()
    coor_df = coor_df.loc[adata.obs_names, ['y', 'x']]
    adata.obsm["spatial"] = coor_df.to_numpy()
    sc.pp.calculate_qc_metrics(adata, inplace=True)
    adata.obs['imagecol'] = coor_df.iloc[:, 1]
    adata.obs['imagerow'] = coor_df.iloc[:, 0]
    return adata

def read_slide_seq(path,
                      library_id=None,
                      scale=None,
                      quality="hires",
                      spot_diameter_fullres=50,
                      background_color="white",):
    count = pd.read_csv(os.path.join(path, "count_matrix.count"))
    meta = pd.read_csv(os.path.join(path, "spatial.idx"))
    adata = AnnData(count.iloc[:, 1:].set_index("gene").T)
    adata.var["ENSEMBL"] = count["ENSEMBL"].values
    adata.obs["index"] = meta["index"].values
    if scale == None:
        max_coor = np.max(meta[["x", "y"]].values)
        scale = 2000 / max_coor
    adata.obs["imagecol"] = meta["x"].values * scale
    adata.obs["imagerow"] = meta["y"].values * scale
    # Create image
    max_size = np.max([adata.obs["imagecol"].max(), adata.obs["imagerow"].max()])
    max_size = int(max_size + 0.1 * max_size)
    if background_color == "black":
        image = Image.new("RGBA", (max_size, max_size), (0, 0, 0, 0))
    else:
        image = Image.new("RGBA", (max_size, max_size), (255, 255, 255, 255))
    imgarr = np.array(image)
    if library_id is None:
        library_id = "Slide-seq"
    adata.uns["spatial"] = {}
    adata.uns["spatial"][library_id] = {}
    adata.uns["spatial"][library_id]["images"] = {}
    adata.uns["spatial"][library_id]["images"][quality] = imgarr
    adata.uns["spatial"][library_id]["use_quality"] = quality
    adata.uns["spatial"][library_id]["scalefactors"] = {}
    adata.uns["spatial"][library_id]["scalefactors"][
        "tissue_" + quality + "_scalef"] = scale
    adata.uns["spatial"][library_id]["scalefactors"][
        "spot_diameter_fullres"
    ] = spot_diameter_fullres
    adata.obsm["spatial"] = meta[["x", "y"]].values
    return adata

def Graph_Construction(adata,args):
    cell_loc = adata.obs[["imagerow","imagecol"]].values
    distance_np = pdist(cell_loc, metric = "euclidean")
    distance_np_X =squareform(distance_np)
    threshold = args.threshold_radius
    num_big = np.where((0< distance_np_X)&(distance_np_X < threshold))[0].shape[0]
    adj_matrix = np.zeros(distance_np_X.shape)
    non_zero_point = np.where((0< distance_np_X)&(distance_np_X<threshold))
    for i in range(num_big):
        x = non_zero_point[0][i]
        y = non_zero_point[1][i]
        adj_matrix[x][y] = 1 
    adj_matrix = adj_matrix + np.eye(distance_np_X.shape[0])
    adj_matrix  = np.float32(adj_matrix)
    adj_matrix_crs = sparse.csr_matrix(adj_matrix)
    graph = dgl.from_scipy(adj_matrix_crs,eweight_name='w')
    adata.var_names=[i.upper() for i in list(adata.var_names)]
    adata.var["genename"]=adata.var.index.astype("str")
    adata.var_names_make_unique
    if(args.feature_dim_method == "PCA"):
        sc.pp.filter_genes(adata, min_cells=5)
        adata_X = sc.pp.normalize_total(adata, target_sum=1, exclude_highly_expressed=True, inplace=False)['X']
        adata_X = sc.pp.scale(adata_X)
        adata_X = sc.pp.pca(adata_X, n_comps=args.num_features)
    else:
        sc.pp.filter_genes(adata, min_cells=5)
        sc.pp.highly_variable_genes(adata, flavor="seurat_v3", n_top_genes=args.num_features)
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        adata_Vars =  adata[:, adata.var['highly_variable']]
        adata_X = adata_Vars.X.todense()
    graph.ndata["feat"] = torch.tensor(adata_X.copy())
    return adata,graph

def Graph_get(adata,threshold = 30,feature_dim_method = "PCA",num_features = 600,row_name = "imagerow",col_name = "imagecol"):
    cell_loc = adata.obs[[row_name,col_name]].values
    distance_np = pdist(cell_loc, metric = "euclidean")
    distance_np_X =squareform(distance_np)
    num_big = np.where((0< distance_np_X)&(distance_np_X < threshold))[0].shape[0]
    adj_matrix = np.zeros(distance_np_X.shape)
    non_zero_point = np.where((0 < distance_np_X) & (distance_np_X < threshold))
    adj_matrix = np.zeros(distance_np_X.shape)
    non_zero_point = np.where((0< distance_np_X)&(distance_np_X<threshold))
    for i in range(num_big):
        x = non_zero_point[0][i]
        y = non_zero_point[1][i]
        adj_matrix[x][y] = 1 
    adj_matrix = adj_matrix + np.eye(distance_np_X.shape[0])
    adj_matrix  = np.float32(adj_matrix)
    adj_matrix_crs = sparse.csr_matrix(adj_matrix)
    graph = dgl.from_scipy(adj_matrix_crs,eweight_name='w')
    if(feature_dim_method == "PCA"):
        sc.pp.normalize_total(adata, target_sum=1, exclude_highly_expressed=True, inplace=False)['X']
        sc.pp.scale(adata)
        sc.pp.pca(adata, n_comps=num_features)
        adata_X = adata.obsm["X_pca"]
    else:
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, flavor='seurat_v3', n_top_genes=num_features)
        adata_Vars =  adata[:, adata.var['highly_variable']]
        adata_X = adata_Vars.X.todense()
    graph.ndata["feat"] = torch.tensor(adata_X.copy())
    print(graph)
    return graph

def GSG_train(adata,graph,args):
    device = args.device if args.device >= 0 else "cpu"
    max_epoch = args.max_epoch
    optim_type = args.optimizer 
    lr = args.lr
    weight_decay = args.weight_decay
    load_model = args.load_model
    args.num_features = args.num_features
    seed = 0
    set_random_seed(seed)
    logger = None
    model = build_model(args)
    model.to(device)
    optimizer = create_optimizer(optim_type, model, lr, weight_decay)
    scheduler = None
    x = graph.ndata["feat"]
    if not load_model:
        model = pretrain(model, graph, x, optimizer, max_epoch, device, scheduler, logger)
    model.train(False)
    x = graph.ndata["feat"]
    embedding = model.embed(graph.to(device), x.to(device))
    adata.obsm["GSG_embedding"] = embedding.cpu().detach().numpy()
    if args.feature_dim_method == "HVG":
        latten_embedding = model.encoder_to_decoder(embedding)
        imputation_embedding =  model.decoder(graph.to(device),latten_embedding)
        adata.obsm["GSG_imputation_embedding"] = imputation_embedding.cpu().detach().numpy()
    return adata,model

def Data_read_and_comb(args,min_cells = 5):
    folder_name = args.folder_name
    before_sample_name = args.before_sample_name
    after_sample_name = args.after_sample_name
    before_folder_name = folder_name + before_sample_name
    adata_before = read_10X_Visium(before_folder_name)
    adata_before.var_names=[i.upper() for i in list(adata_before.var_names)]
    adata_before.var["genename"]=adata_before.var.index.astype("str")
    adata_before.var_names_make_unique()
    sc.pp.filter_genes(adata_before, min_cells)
    after_folder_name = folder_name + after_sample_name
    adata_after = read_10X_Visium(after_folder_name)
    adata_after.var_names=[i.upper() for i in list(adata_after.var_names)]
    adata_after.var["genename"]=adata_after.var.index.astype("str")
    adata_after.var_names_make_unique()
    sc.pp.filter_genes(adata_after, min_cells)
    adata_before.obs_names_make_unique()
    adata_after.obs_names_make_unique()
    adata_before.var_names_make_unique()
    adata_after.var_names_make_unique()
    adata_before.obs["sample_number"] = before_sample_name
    adata_after.obs["sample_number"] = after_sample_name
    merge_gene_list = list(set(adata_before.var.index).intersection(set(adata_after.var.index)))
    adata_before.var["comb_gene"] = True
    for i in adata_before.var.index:
        if i not in merge_gene_list:
            adata_before.var.loc[i,"comb_gene"] = False
    adata_after.var["comb_gene"] = True
    for i in adata_after.var.index:
        if i not in merge_gene_list:
            adata_after.var.loc[i,"comb_gene"] = False
    adata_before.obs.index=['sub_before-'+x for x in adata_before.obs.index]
    adata_before.obs.index
    adata_after.obs.index=['sub_after-'+x for x in adata_after.obs.index]
    adata_after.obs.index
    adata_subset_before = adata_before[:,adata_before.var['comb_gene']]
    adata_subset_after = adata_after[:,adata_after.var['comb_gene']]
    adatas=[adata_subset_before,adata_subset_after]
    adatas = ad.concat(adatas,merge= "same")
    return adatas

def Graph_Get_And_Data_Charge(adatas,args):
    feature_dim_method = args.feature_dim_method
    num_features = args.num_features
    before_sample_name = args.before_sample_name
    after_sample_name = args.after_sample_name
    adata_before = adatas[adatas.obs.sample_number == before_sample_name,]
    adata_after = adatas[adatas.obs.sample_number ==  after_sample_name,]
    graph_before = Graph_get(adata=adata_before,feature_dim_method = feature_dim_method , num_features= num_features, threshold =30)
    num_features = graph_before.ndata["feat"].shape[1]
    graph_after = Graph_get(adata=adata_after,feature_dim_method = feature_dim_method , num_features= num_features, threshold = 30)
    num_features = graph_after.ndata["feat"].shape[1]
    sc.pp.neighbors(adatas, n_neighbors=10, n_pcs=100,use_rep='X')
    sc.pp.normalize_total(adatas, target_sum=1, exclude_highly_expressed=True, inplace=False)['X']
    sc.pp.scale(adatas)
    sc.pp.pca(adatas, n_comps=num_features)
    two_graph = [graph_before,graph_after]
    two_adata = [adata_before,adata_after]
    return two_graph,two_adata

def Origin_PCA_show(adatas,result_file,args):
    before_sample_name = args.before_sample_name
    after_sample_name = args.after_sample_name
    sc.pp.neighbors(adatas, n_neighbors=10, n_pcs=100,use_rep='X')
    fig, ax = plt.subplots(2,1,figsize=(10, 20))
    sc.tl.umap(adatas)
    sc.pl.umap(adatas, color="sample_number",title = "comb_umap_Origin",show = False ,ax=ax[0])
    sc.pp.neighbors(adatas, n_neighbors=10, n_pcs=600)
    sc.tl.umap(adatas)
    sc.pl.umap(adatas, color="sample_number",title = "comb_umap_PCA",show = False ,ax=ax[1])
    fig.savefig(result_file + "/Origin_and_PCA_umap.pdf",bbox_inches = 'tight')

def GSG_batch_effect_train(args,two_graph):
    device = args.device if args.device >= 0 else "cpu"
    max_epoch = args.max_epoch
    optim_type = args.optimizer 
    lr = args.lr
    weight_decay = args.weight_decay
    seed = 0
    set_random_seed(seed)
    model = build_model(args)
    graph_before = two_graph[0]
    graph_after = two_graph[1]
    model.to(device)
    x_after = graph_after.ndata["feat"]
    x_before = graph_before.ndata["feat"]
    graph_before = graph_before.to(device)
    x_before = x_before.to(device)
    graph_after = graph_after.to(device)
    x_after = x_after.to(device)
    optimizer = create_optimizer(optim_type, model, lr, weight_decay)
    epoch_iter = tqdm(range(max_epoch))
    for epoch in epoch_iter:
        model.train()
        loss, loss_dict = model(graph_before, x_before)
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        model.train()
        epoch_iter.set_description(f"# Epoch {epoch}: train_loss: {loss.item():.4f}")
    return model

def GSG_Embedding_Get(model,two_graph,two_adata,args):
    device = args.device if args.device >= 0 else "cpu"
    graph_before = two_graph[0]
    graph_after = two_graph[1]
    adata_before = two_adata[0]
    adata_after = two_adata[1]
    x = graph_after.ndata["feat"]
    embedding = model.embed(graph_after.to(device), x.to(device))
    adata_after.obsm["GSG_embedding"] = embedding.cpu().detach().numpy()
    if args.feature_dim_method == "HVG":
        latten_embedding = model.encoder_to_decoder(embedding)
        imputation_embedding =  model.decoder(graph_after.to(device),latten_embedding)
        adata_after.obsm["GSG_imputation_embedding"] = imputation_embedding.cpu().detach().numpy()
    x = graph_before.ndata["feat"]
    embedding = model.embed(graph_before.to(device), x.to(device))
    adata_before.obsm["GSG_embedding"] = embedding.cpu().detach().numpy()
    if args.feature_dim_method == "HVG":
        latten_embedding = model.encoder_to_decoder(embedding)
        imputation_embedding =  model.decoder(graph_before.to(device),latten_embedding)
        adata_before.obsm["GSG_imputation_embedding"] = imputation_embedding.cpu().detach().numpy()

def GSG_Cluster_Comb(two_adata,args,result_file):
    color_list = {
        "0" : "#E41A1C",
        "1": "#377EB8",
        "2":"#4DAF4A",
        "3":"#984EA3",
        "4":"#FF7F00",
        "5":"#FFFF33",
        "6":"#A65628",
        "7" :"#F781BF",
        "8":"#999999",
        "9":"#E41A80",
        "10":"#377F1C",
        "11":"#4DAFAE",
        "12":"#984F07",
        "13":"#FF7F64",
        "14":"#FFFF97",
        "15":"#A6568C", 
        "16":"#F78223", 
        "17":"#9999FD", 
        "18" :"#E41AE4", 
        "19" : "#377F80", 
        "20":"#4DB012",
        "21":"#984F6B",
        "22":"#FF7FC8",  
        "23":"#A656F0", 
        "24":"#F78287", 
        "25":"#999A61", 
        "26":"#E41B48", 
        "27":"#377FE4",
        "28":"#4DB076", 
        "29":"#984FCF", 
        "30":"#FF802C",
    }
    flatui = ["#E41A1C", "#377EB8", "#4DAF4A", "#984EA3", "#FF7F00", "#FFFF33", "#A65628", "#F781BF", "#999999", "#E41A80",
    "#377F1C", "#4DAFAE", "#984F07", "#FF7F64", "#FFFF97", "#A6568C", "#F78223", "#9999FD", "#E41AE4", "#377F80", "#4DB012",
    "#984F6B", "#FF7FC8",  "#A656F0", "#F78287", "#999A61", "#E41B48", "#377FE4", "#4DB076", "#984FCF", "#FF802C"]
    import seaborn as sns
    adata_before = two_adata[0]
    adata_after = two_adata[1]
    adatas_GSG =[adata_before,adata_after]
    adatas_GSG = ad.concat(adatas_GSG,merge = "same")
    K_means_value = KMeans_use(adatas_GSG.obsm['GSG_embedding'],args.num_classes)
    adatas_GSG.obs["Comb_K_means_Number"] = K_means_value
    sc.pp.neighbors(adatas_GSG, n_neighbors=10,use_rep='GSG_embedding')
    sc.tl.umap(adatas_GSG)
    fig_new = sc.pl.umap(adatas_GSG, color="sample_number",title = "GSG_Comb_UMAP",return_fig = True)
    fig_new.savefig(result_file + "/GSG_Comb_UMAP.pdf",bbox_inches = 'tight')
    adatas_GSG.obs["Comb_K_means_Number_str"] = adatas_GSG.obs["Comb_K_means_Number"].astype("str")
    fig_new = sc.pl.umap(adatas_GSG, color="Comb_K_means_Number_str",title = "GSG_Kmeans_Comb_Cluster_UMAP",return_fig = True,palette=sns.color_palette(flatui))
    fig_new.savefig(result_file + "/GSG_Kmeans_Comb_Cluster_UMAP.pdf",bbox_inches = 'tight')
    adata_before = adatas_GSG[adatas_GSG.obs.sample_number == args.before_sample_name,]
    adata_after = adatas_GSG[adatas_GSG.obs.sample_number == args.after_sample_name,]
    adata_before.obs["cluster_color"] = ""
    adata_after.obs["cluster_color"] = ""
    for i in adata_before.obs.index:
        adata_before.obs.loc[i,"cluster_color"] = color_list[adata_before.obs.loc[i,"Comb_K_means_Number_str"]]
    for i in adata_after.obs.index:
        adata_after.obs.loc[i,"cluster_color"] = color_list[adata_after.obs.loc[i,"Comb_K_means_Number_str"]]
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go
    col_name ="imagecol"
    row_name = "imagerow"
    fig = make_subplots(rows=1, cols=2)
    fig.add_trace(go.Scatter(x=adata_before.obs[col_name], y=adata_before.obs[row_name],
                        mode='markers',
                        name='markers',
                        marker=dict(size=14,color = adata_before.obs["cluster_color"]),
                    ),row=1, col=1)
    fig.add_trace(go.Scatter(x=adata_after.obs[col_name], y= adata_after.obs[row_name],
                        mode='markers',
                        name='markers',
                        marker=dict(size=14,color = adata_after.obs["cluster_color"]),
                    ),row=1, col=2)
    fig.update_layout(height=1000, width=2000, title_text="batch_show")
    fig.show()
    fig.write_image(result_file + "/" + args.before_sample_name + "_" +args.after_sample_name+ "_spatial_comb_kmeans.pdf")
    return adatas_GSG

def GSG_Kmean_Separate_cluster(adatas_GSG,args):
    adata_before = adatas_GSG[adatas_GSG.obs.sample_number == args.before_sample_name,]
    adata_after = adatas_GSG[adatas_GSG.obs.sample_number == args.after_sample_name,]
    k_means_value = KMeans_use(adata_after.obsm['GSG_embedding'],args.num_classes)
    adata_after.obs["GSG_Kmeans_Cluster_Separate_number"] = k_means_value
    k_means_value = KMeans_use(adata_before.obsm['GSG_embedding'],args.num_classes)
    adata_before.obs["GSG_Kmeans_Cluster_Separate_number"] = k_means_value
    adatas_GSG =[adata_before,adata_after]
    adatas_GSG = ad.concat(adatas_GSG,merge = "same")
    return adatas_GSG

def GSG_KMean_cluster_Separate_Result_Show(adatas_GSG,args,result_file):
    flatui = ["#E41A1C", "#377EB8", "#4DAF4A", "#984EA3", "#FF7F00", "#FFFF33", "#A65628", "#F781BF", "#999999", "#E41A80",
    "#377F1C", "#4DAFAE", "#984F07", "#FF7F64", "#FFFF97", "#A6568C", "#F78223", "#9999FD", "#E41AE4", "#377F80", "#4DB012",
    "#984F6B", "#FF7FC8",  "#A656F0", "#F78287", "#999A61", "#E41B48", "#377FE4", "#4DB076", "#984FCF", "#FF802C"]
    sc.pp.neighbors(adatas_GSG, n_neighbors=10,use_rep='GSG_embedding')
    sc.tl.umap(adatas_GSG)
    fig_new = sc.pl.umap(adatas_GSG, color="sample_number",title = "GSG_Comb_UMAP",return_fig = True)
    fig_new.savefig(result_file + "/GSG_UMAP.pdf",bbox_inches = 'tight')
    adatas_GSG.obs["GSG_Kmeans_Cluster_Separate_number_str"] = adatas_GSG.obs["GSG_Kmeans_Cluster_Separate_number"].astype("str")
    adata_before = adatas_GSG[adatas_GSG.obs.sample_number == args.before_sample_name,]
    adata_after = adatas_GSG[adatas_GSG.obs.sample_number == args.after_sample_name,]
    GSG_adata_before_cluster_file = result_file + "/" + args.before_sample_name + "_Separate_Cluster.pdf"
    drawPicture(adata_before.obs,col_name ="imagecol",row_name = "imagerow",colorattribute="GSG_Kmeans_Cluster_Separate_number_str",save_file = GSG_adata_before_cluster_file,is_show=False,is_save= True)
    GSG_adata_after_cluster_file = result_file + "/" + args.after_sample_name + "_Separate_Cluster.pdf"
    drawPicture(adata_after.obs,col_name ="imagecol",row_name = "imagerow",colorattribute="GSG_Kmeans_Cluster_Separate_number_str",save_file = GSG_adata_after_cluster_file,is_show=False,is_save= True)
    return adatas_GSG

def GSG_plot_batch_effect_imputation(adata,args,result_file):
    diff_gene_floder = result_file + "/differential_genes/"
    mkdir(diff_gene_floder)
    sc.pp.filter_cells(adata, min_genes=200)
    sc.pp.filter_genes(adata, min_cells=3)
    sc.pp.pca(adata, n_comps=600)
    adata.obs["Comb_k_means_number_str"] = adata.obs["Comb_K_means_Number_str"].astype("category")
    adata.var_names = adata.var.index
    sc.tl.rank_genes_groups(adata, "Comb_k_means_number_str" , use_raw=False,n_genes = 300,method="wilcoxon")
    sc.tl.filter_rank_genes_groups(adata)
    celltype_list = adata.uns["rank_genes_groups_filtered"]["names"].dtype.names
    for celltype in celltype_list:
        gene_list = []
        for gene in adata.uns["rank_genes_groups_filtered"]["names"][celltype]:
            if(type(gene) == str):
                gene_list.append(gene)
            if(len(gene_list) >= 3):
                break
        if(len(gene_list) != 0 ):
            sc.pl.violin(adata, gene_list, groupby='Comb_k_means_number_str',show=False)
            plt.savefig(diff_gene_floder + celltype+ "_gene_violin.pdf")
    adata.var['features'] = adata.var.index
    result_csv = pd.DataFrame()
    gene_num = 0
    celltype_list = adata.uns["rank_genes_groups_filtered"]["names"].dtype.names
    for celltype in celltype_list:
        gene_list = adata.uns["rank_genes_groups_filtered"]["names"][celltype]
        for gene in gene_list:
            if(type(gene) == str):
                result_csv.loc[gene_num,"gene_num"] = gene
                result_csv.loc[gene_num,"gene"] = adata.var.loc[gene,"features"]
                result_csv.loc[gene_num,"celltype"] = celltype
                result_csv.loc[gene_num,"type"] = "after"
                gene_num = gene_num + 1
                print(gene)
                print(celltype)
    result_csv.to_csv(diff_gene_floder + "all_differential_genes_result.csv")
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go
    import plotly.express as px
    col_name ="imagecol"
    row_name = "imagerow"    
    adata_before = adata[adata.obs.sample_number == args.before_sample_name,]
    before_cell_csv = adata_before.obs[[row_name,col_name]]
    before_celltype_list = adata_before.uns["rank_genes_groups"]["names"].dtype.names
    sc.pp.normalize_total(adata_before, target_sum=1e4)
    sc.pp.log1p(adata_before)
    sc.pp.highly_variable_genes(adata_before, flavor='seurat_v3', n_top_genes=args.num_features)
    before_adata_Vars =  adata_before[:, adata_before.var['highly_variable']]
    for celltype in before_celltype_list:
        gene_list = adata_before.uns["rank_genes_groups"]["names"][celltype][:5]
        for gene in gene_list:
            if(gene in before_adata_Vars.var.index):
                print(gene)
                print(celltype)
                celltype = celltype.replace("/","_")
                before_gene_csv = pd.DataFrame(index =adata_before.obs.index,columns= adata_before.var.index,data =adata_before.X.todense())
                after_gene_csv = pd.DataFrame(index =before_adata_Vars.obs.index,columns= before_adata_Vars.var.index,data =adata_before.obsm["GSG_imputation_embedding"] )
                before_gene_csv_merge = pd.merge(before_gene_csv,before_cell_csv,left_index=True,right_index=True)
                after_gene_csv_merge = pd.merge(after_gene_csv,before_cell_csv,left_index=True,right_index=True)
                width = 1000
                height = 2000
                marker_size = 10
                fig = make_subplots(rows=1, cols=2)
                fig.add_trace(go.Scatter(x=before_gene_csv_merge[col_name], y=before_gene_csv_merge[row_name],
                                    mode='markers',
                                    name='markers',
                                    marker=dict(size=10,color = before_gene_csv_merge[gene],colorscale = px.colors.sequential.Viridis),
                                ),row=1, col=1)
                fig.add_trace(go.Scatter(x=after_gene_csv_merge[col_name], y=after_gene_csv_merge[row_name],
                                    mode='markers',
                                    name='markers',
                                    marker=dict(size=10,color = after_gene_csv_merge[gene],colorscale = px.colors.sequential.Viridis),
                                ),row=1, col=2)
                fig.update_layout(height=1000, width=2000, title_text="before and after " + gene + " "+ celltype + " gene express ")
                fig.write_image(diff_gene_floder + "first_sample_before_and_after_" + gene + "_"+ celltype + "_gene express.pdf")
    adata_after = adata[adata.obs.sample_number == args.after_sample_name,]
    after_cell_csv = adata_after.obs[[row_name,col_name]]
    after_celltype_list = adata_after.uns["rank_genes_groups"]["names"].dtype.names
    sc.pp.normalize_total(adata_after, target_sum=1e4)
    sc.pp.log1p(adata_after)
    sc.pp.highly_variable_genes(adata_after, flavor='seurat_v3', n_top_genes=args.num_features)
    after_adata_Vars =  adata_after[:, adata_after.var['highly_variable']]
    for celltype in after_celltype_list:
        gene_list = adata_after.uns["rank_genes_groups"]["names"][celltype][:5]
        for gene in gene_list:
            if(gene in after_adata_Vars.var.index):
                print(gene)
                print(celltype)
                celltype = celltype.replace("/","_")
                before_gene_csv = pd.DataFrame(index =adata_after.obs.index,columns= adata_after.var.index,data =adata_after.X.todense())
                after_gene_csv = pd.DataFrame(index =after_adata_Vars.obs.index,columns= after_adata_Vars.var.index,data =adata_after.obsm["GSG_imputation_embedding"] )
                before_gene_csv_merge = pd.merge(before_gene_csv,after_cell_csv,left_index=True,right_index=True)
                after_gene_csv_merge = pd.merge(after_gene_csv,after_cell_csv,left_index=True,right_index=True)
                width = 1000
                height = 2000
                marker_size = 10
                fig = make_subplots(rows=1, cols=2)
                fig.add_trace(go.Scatter(x=before_gene_csv_merge[col_name], y=before_gene_csv_merge[row_name],
                                    mode='markers',
                                    name='markers',
                                    marker=dict(size=10,color = before_gene_csv_merge[gene],colorscale = px.colors.sequential.Viridis),
                                ),row=1, col=1)
                fig.add_trace(go.Scatter(x=after_gene_csv_merge[col_name], y=after_gene_csv_merge[row_name],
                                    mode='markers',
                                    name='markers',
                                    marker=dict(size=10,color = after_gene_csv_merge[gene],colorscale = px.colors.sequential.Viridis),
                                ),row=1, col=2)
                fig.update_layout(height=1000, width=2000, title_text="before and after " + gene + " "+ celltype + " gene express ")
                fig.write_image(diff_gene_floder + "second_sample_before_and_after_" + gene + "_"+ celltype + "_gene express.pdf")
    del adata.uns['rank_genes_groups_filtered']

def GSG_plot_imputation_gene(adata,args,result_file):
    diff_gene_floder = result_file + "/" + args.sample_name + "_differential_genes/"
    mkdir(diff_gene_floder)
    sc.pp.filter_cells(adata, min_genes=200)
    sc.pp.filter_genes(adata, min_cells=3)
    sc.pp.pca(adata, n_comps=600)
    adata.obs["GSG_kmeans_cluster_str"] =  adata.obs["GSG_Kmeans_cluster_str"].astype("category")
    adata.var_names = adata.var.index
    sc.tl.rank_genes_groups(adata, "GSG_kmeans_cluster_str" , use_raw=False,n_genes = 300,method="wilcoxon")
    sc.tl.filter_rank_genes_groups(adata)
    celltype_list = adata.uns["rank_genes_groups_filtered"]["names"].dtype.names
    for celltype in celltype_list:
        gene_list = []
        for gene in adata.uns["rank_genes_groups_filtered"]["names"][celltype]:
            if(type(gene) == str):
                gene_list.append(gene)
            if(len(gene_list) >= 3):
                break
        if(len(gene_list) != 0 ):
            sc.pl.violin(adata, gene_list, groupby='GSG_kmeans_cluster_str',show=False)
            plt.savefig(diff_gene_floder + celltype+ "_gene_violin.pdf")
    adata.var['features'] = adata.var.index
    result_csv = pd.DataFrame()
    gene_num = 0
    celltype_list = adata.uns["rank_genes_groups_filtered"]["names"].dtype.names
    for celltype in celltype_list:
        gene_list = adata.uns["rank_genes_groups_filtered"]["names"][celltype]
        for gene in gene_list:
            if(type(gene) == str):
                result_csv.loc[gene_num,"gene_num"] = gene
                result_csv.loc[gene_num,"gene"] = adata.var.loc[gene,"features"]
                result_csv.loc[gene_num,"celltype"] = celltype
                result_csv.loc[gene_num,"type"] = "after"
                gene_num = gene_num + 1
    result_csv.to_csv(diff_gene_floder + "all_differential_genes_result.csv")
    from plotly.subplots import make_subplots
    import plotly.graph_objects as go
    import plotly.express as px
    col_name ="imagecol"
    row_name = "imagerow"
    cell_csv = adata.obs[[row_name,col_name]]
    celltype_list = adata.uns["rank_genes_groups"]["names"].dtype.names
    adata_Vars =  adata[:, adata.var['highly_variable']]
    for celltype in celltype_list:
        gene_list = adata.uns["rank_genes_groups"]["names"][celltype][:5]
        for gene in gene_list:
            if(gene in adata_Vars.var.index):
                print(gene)
                print(celltype)
                celltype = celltype.replace("/","_")
                before_gene_csv = pd.DataFrame(index =adata.obs.index,columns= adata.var.index,data =adata.X.todense())
                after_gene_csv = pd.DataFrame(index =adata_Vars.obs.index,columns= adata_Vars.var.index,data =adata.obsm["GSG_imputation_embedding"] )
                before_gene_csv_merge = pd.merge(before_gene_csv,cell_csv,left_index=True,right_index=True)
                after_gene_csv_merge = pd.merge(after_gene_csv,cell_csv,left_index=True,right_index=True)
                width = 1000
                height = 2000
                marker_size = 10
                fig = make_subplots(rows=1, cols=2)
                fig.add_trace(go.Scatter(x=before_gene_csv_merge[col_name], y=before_gene_csv_merge[row_name],
                                    mode='markers',
                                    name='markers',
                                    marker=dict(size=10,color = before_gene_csv_merge[gene],colorscale = px.colors.sequential.Viridis),
                                ),row=1, col=1)
                fig.add_trace(go.Scatter(x=after_gene_csv_merge[col_name], y=after_gene_csv_merge[row_name],
                                    mode='markers',
                                    name='markers',
                                    marker=dict(size=10,color = after_gene_csv_merge[gene],colorscale = px.colors.sequential.Viridis),
                                ),row=1, col=2)
                fig.update_layout(height=1000, width=2000, title_text="before and after " + gene + " "+ celltype + " gene express ")
                fig.write_image(diff_gene_floder + "before_and_after_" + gene + "_"+ celltype + "_gene express.pdf")
    del adata.uns['rank_genes_groups_filtered']

def GSG_plot_trajectory(adata, args, result_file, filename = "GSG_Cluster_Embedding_trajectory.pdf"):
    if(args.cluster_label != ""):
        used_adata = adata[adata.obs[args.cluster_label]!= "None",]
        sc.tl.paga(used_adata, groups=args.cluster_label)
    else:
        used_adata = adata
        sc.tl.paga(used_adata, groups="GSG_Kmeans_cluster_str")
    sc.pl.paga_compare(used_adata, legend_fontsize=15, node_size_scale= 4, frameon=False, size=50,max_edge_width = 10,right_margin=0.2,
                        legend_fontoutline=5,title="PAGA_SHOW", show=False)
    plt.savefig( result_file + "/" + filename,bbox_inches='tight',dpi =1000)


# def save_file(adata,result_file):
#     adata_X = adata.X



import torch
import torch.nn as nn
import torch.nn.functional as F

import dgl.function as fn
from dgl.utils import expand_as_pair

from tools.utils import create_activation, create_norm


class GIN(nn.Module):
    def __init__(self,
                 in_dim,
                 num_hidden,
                 out_dim,
                 num_layers,
                 dropout,
                 activation,
                 residual,
                 norm,
                 encoding=False,
                 learn_eps=False,
                 aggr="sum",
                 ):
        super(GIN, self).__init__()
        self.out_dim = out_dim
        self.num_layers = num_layers
        self.layers = nn.ModuleList()
        self.activation = activation
        self.dropout = dropout

        last_activation = create_activation(activation) if encoding else None
        last_residual = encoding and residual
        last_norm = norm if encoding else None
        
        if num_layers == 1:
            apply_func = MLP(2, in_dim, num_hidden, out_dim, activation=activation, norm=norm)
            if last_norm:
                apply_func = ApplyNodeFunc(apply_func, norm=norm, activation=activation)
            self.layers.append(GINConv(in_dim, out_dim, apply_func, init_eps=0, learn_eps=learn_eps, residual=last_residual))
        else:
            # input projection (no residual)
            self.layers.append(GINConv(
                in_dim, 
                num_hidden, 
                ApplyNodeFunc(MLP(2, in_dim, num_hidden, num_hidden, activation=activation, norm=norm), activation=activation, norm=norm), 
                init_eps=0,
                learn_eps=learn_eps,
                residual=residual)
                )
            # hidden layers
            for l in range(1, num_layers - 1):
                # due to multi-head, the in_dim = num_hidden * num_heads
                self.layers.append(GINConv(
                    num_hidden, num_hidden, 
                    ApplyNodeFunc(MLP(2, num_hidden, num_hidden, num_hidden, activation=activation, norm=norm), activation=activation, norm=norm), 
                    init_eps=0,
                    learn_eps=learn_eps,
                    residual=residual)
                )
            # output projection
            apply_func = MLP(2, num_hidden, num_hidden, out_dim, activation=activation, norm=norm)
            if last_norm:
                apply_func = ApplyNodeFunc(apply_func, activation=activation, norm=norm)

            self.layers.append(GINConv(num_hidden, out_dim, apply_func, init_eps=0, learn_eps=learn_eps, residual=last_residual))

        self.head = nn.Identity()

    def forward(self, g, inputs, return_hidden=False):
        h = inputs
        hidden_list = []
        for l in range(self.num_layers):
            h = F.dropout(h, p=self.dropout, training=self.training)
            h = self.layers[l](g, h)
            hidden_list.append(h)
        # output projection
        if return_hidden:
            return self.head(h), hidden_list
        else:
            return self.head(h)

    def reset_classifier(self, num_classes):
        self.head = nn.Linear(self.out_dim, num_classes)


class GINConv(nn.Module):
    def __init__(self,
                 in_dim,
                 out_dim,
                 apply_func,
                 aggregator_type="sum",
                 init_eps=0,
                 learn_eps=False,
                 residual=False,
                 ):
        super().__init__()
        self._in_feats = in_dim
        self._out_feats = out_dim
        self.apply_func = apply_func

        self._aggregator_type = aggregator_type
        if aggregator_type == 'sum':
            self._reducer = fn.sum
        elif aggregator_type == 'max':
            self._reducer = fn.max
        elif aggregator_type == 'mean':
            self._reducer = fn.mean
        else:
            raise KeyError('Aggregator type {} not recognized.'.format(aggregator_type))
            
        if learn_eps:
            self.eps = torch.nn.Parameter(torch.FloatTensor([init_eps]))
        else:
            self.register_buffer('eps', torch.FloatTensor([init_eps]))

        if residual:
            if self._in_feats != self._out_feats:
                self.res_fc = nn.Linear(
                    self._in_feats, self._out_feats, bias=False)
                print("! Linear Residual !")
            else:
                print("Identity Residual ")
                self.res_fc = nn.Identity()
        else:
            self.register_buffer('res_fc', None)

    def forward(self, graph, feat):
        with graph.local_scope():
            aggregate_fn = fn.copy_src('h', 'm')

            feat_src, feat_dst = expand_as_pair(feat, graph)
            graph.srcdata['h'] = feat_src
            graph.update_all(aggregate_fn, self._reducer('m', 'neigh'))
            rst = (1 + self.eps) * feat_dst + graph.dstdata['neigh']
            if self.apply_func is not None:
                rst = self.apply_func(rst)

            if self.res_fc is not None:
                rst = rst + self.res_fc(feat_dst)

            return rst


class ApplyNodeFunc(nn.Module):
    """Update the node feature hv with MLP, BN and ReLU."""
    def __init__(self, mlp, norm="batchnorm", activation="relu"):
        super(ApplyNodeFunc, self).__init__()
        self.mlp = mlp
        norm_func = create_norm(norm)
        if norm_func is None:
            self.norm = nn.Identity()
        else:
            self.norm = norm_func(self.mlp.output_dim)
        self.act = create_activation(activation)

    def forward(self, h):
        h = self.mlp(h)
        h = self.norm(h)
        h = self.act(h)
        return h


class MLP(nn.Module):
    """MLP with linear output"""
    def __init__(self, num_layers, input_dim, hidden_dim, output_dim, activation="relu", norm="batchnorm"):
        super(MLP, self).__init__()
        self.linear_or_not = True  # default is linear model
        self.num_layers = num_layers
        self.output_dim = output_dim

        if num_layers < 1:
            raise ValueError("number of layers should be positive!")
        elif num_layers == 1:
            # Linear model
            self.linear = nn.Linear(input_dim, output_dim)
        else:
            # Multi-layer model
            self.linear_or_not = False
            self.linears = torch.nn.ModuleList()
            self.norms = torch.nn.ModuleList()
            self.activations = torch.nn.ModuleList()

            self.linears.append(nn.Linear(input_dim, hidden_dim))
            for layer in range(num_layers - 2):
                self.linears.append(nn.Linear(hidden_dim, hidden_dim))
            self.linears.append(nn.Linear(hidden_dim, output_dim))

            for layer in range(num_layers - 1):
                self.norms.append(create_norm(norm)(hidden_dim))
                self.activations.append(create_activation(activation))

    def forward(self, x):
        if self.linear_or_not:
            # If linear model
            return self.linear(x)
        else:
            # If MLP
            h = x
            for i in range(self.num_layers - 1):
                h = self.norms[i](self.linears[i](h))
                h = self.activations[i](h)
            return self.linears[-1](h)


import torch
import torch.nn as nn

from dgl.ops import edge_softmax
import dgl.function as fn
from dgl.utils import expand_as_pair

from tools.utils import create_activation


class GAT(nn.Module):
    def __init__(self,
                 in_dim,
                 num_hidden,
                 out_dim,
                 num_layers,
                 nhead,
                 nhead_out,
                 activation,
                 feat_drop,
                 attn_drop,
                 negative_slope,
                 residual,
                 norm,
                 concat_out=False,
                 encoding=False
                 ):
        super(GAT, self).__init__()
        self.out_dim = out_dim
        self.num_heads = nhead
        self.num_layers = num_layers
        self.gat_layers = nn.ModuleList()
        self.activation = activation
        self.concat_out = concat_out

        last_activation = create_activation(activation) if encoding else None
        last_residual = (encoding and residual)
        last_norm = norm if encoding else None
        
        if num_layers == 1:
            self.gat_layers.append(GATConv(
                in_dim, out_dim, nhead_out,
                feat_drop, attn_drop, negative_slope, last_residual, norm=last_norm, concat_out=concat_out))
        else:
            # input projection (no residual)
            self.gat_layers.append(GATConv(
                in_dim, num_hidden, nhead,
                feat_drop, attn_drop, negative_slope, residual, create_activation(activation), norm=norm, concat_out=concat_out))
            # hidden layers
            for l in range(1, num_layers - 1):
                # due to multi-head, the in_dim = num_hidden * num_heads
                self.gat_layers.append(GATConv(
                    num_hidden * nhead, num_hidden, nhead,
                    feat_drop, attn_drop, negative_slope, residual, create_activation(activation), norm=norm, concat_out=concat_out))
            # output projection
            self.gat_layers.append(GATConv(
                num_hidden * nhead, out_dim, nhead_out,
                feat_drop, attn_drop, negative_slope, last_residual, activation=last_activation, norm=last_norm, concat_out=concat_out))

        # if norm is not None:
        #     self.norms = nn.ModuleList([
        #         norm(num_hidden * nhead)
        #         for _ in range(num_layers - 1)
        #     ])
        #     if self.concat_out:
        #         self.norms.append(norm(num_hidden * nhead))
        # else:
        #     self.norms = None
    
        self.head = nn.Identity()

    # def forward(self, g, inputs):
    #     h = inputs
    #     for l in range(self.num_layers):
    #         h = self.gat_layers[l](g, h)
    #         if l != self.num_layers - 1:
    #             h = h.flatten(1)
    #             if self.norms is not None:
    #                 h = self.norms[l](h)
    #     # output projection
    #     if self.concat_out:
    #         out = h.flatten(1)
    #         if self.norms is not None:
    #             out = self.norms[-1](out)
    #     else:
    #         out = h.mean(1)
    #     return self.head(out)
    
    def forward(self, g, inputs, return_hidden=False):
        h = inputs
        hidden_list = []
        for l in range(self.num_layers):
            h = self.gat_layers[l](g, h)
            hidden_list.append(h)
            # h = h.flatten(1)
        # output projection
        if return_hidden:
            return self.head(h), hidden_list
        else:
            return self.head(h)

    def reset_classifier(self, num_classes):
        self.head = nn.Linear(self.num_heads * self.out_dim, num_classes)



class GATConv(nn.Module):
    def __init__(self,
                 in_feats,
                 out_feats,
                 num_heads,
                 feat_drop=0.,
                 attn_drop=0.,
                 negative_slope=0.2,
                 residual=False,
                 activation=None,
                 allow_zero_in_degree=False,
                 bias=True,
                 norm=None,
                 concat_out=True):
        super(GATConv, self).__init__()
        self._num_heads = num_heads
        self._in_src_feats, self._in_dst_feats = expand_as_pair(in_feats)
        self._out_feats = out_feats
        self._allow_zero_in_degree = allow_zero_in_degree
        self._concat_out = concat_out

        if isinstance(in_feats, tuple):
            self.fc_src = nn.Linear(
                self._in_src_feats, out_feats * num_heads, bias=False)
            self.fc_dst = nn.Linear(
                self._in_dst_feats, out_feats * num_heads, bias=False)
        else:
            self.fc = nn.Linear(
                self._in_src_feats, out_feats * num_heads, bias=False)
        self.attn_l = nn.Parameter(torch.FloatTensor(size=(1, num_heads, out_feats)))
        self.attn_r = nn.Parameter(torch.FloatTensor(size=(1, num_heads, out_feats)))
        self.feat_drop = nn.Dropout(feat_drop)
        self.attn_drop = nn.Dropout(attn_drop)
        self.leaky_relu = nn.LeakyReLU(negative_slope)
        if bias:
            self.bias = nn.Parameter(torch.FloatTensor(size=(num_heads * out_feats,)))
        else:
            self.register_buffer('bias', None)
        if residual:
            if self._in_dst_feats != out_feats * num_heads:
                self.res_fc = nn.Linear(
                    self._in_dst_feats, num_heads * out_feats, bias=False)
            else:
                self.res_fc = nn.Identity()
        else:
            self.register_buffer('res_fc', None)
        self.reset_parameters()
        self.activation = activation
        # if norm is not None:
        #     self.norm = norm(num_heads * out_feats)
        # else:
        #     self.norm = None
    
        self.norm = norm
        if norm is not None:
            self.norm = norm(num_heads * out_feats)

    def reset_parameters(self):
        """

        Description
        -----------
        Reinitialize learnable parameters.

        Note
        ----
        The fc weights :math:`W^{(l)}` are initialized using Glorot uniform initialization.
        The attention weights are using xavier initialization method.
        """
        gain = nn.init.calculate_gain('relu')
        if hasattr(self, 'fc'):
            nn.init.xavier_normal_(self.fc.weight, gain=gain)
        else:
            nn.init.xavier_normal_(self.fc_src.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_dst.weight, gain=gain)
        nn.init.xavier_normal_(self.attn_l, gain=gain)
        nn.init.xavier_normal_(self.attn_r, gain=gain)
        if self.bias is not None:
            nn.init.constant_(self.bias, 0)
        if isinstance(self.res_fc, nn.Linear):
            nn.init.xavier_normal_(self.res_fc.weight, gain=gain)

    def set_allow_zero_in_degree(self, set_value):
        self._allow_zero_in_degree = set_value

    def forward(self, graph, feat, get_attention=False):
        with graph.local_scope():
            if not self._allow_zero_in_degree:
                if (graph.in_degrees() == 0).any():
                    raise RuntimeError('There are 0-in-degree nodes in the graph, '
                                   'output for those nodes will be invalid. '
                                   'This is harmful for some applications, '
                                   'causing silent performance regression. '
                                   'Adding self-loop on the input graph by '
                                   'calling `g = dgl.add_self_loop(g)` will resolve '
                                   'the issue. Setting ``allow_zero_in_degree`` '
                                   'to be `True` when constructing this module will '
                                   'suppress the check and let the code run.')

            if isinstance(feat, tuple):
                src_prefix_shape = feat[0].shape[:-1]
                dst_prefix_shape = feat[1].shape[:-1]
                h_src = self.feat_drop(feat[0])
                h_dst = self.feat_drop(feat[1])
                if not hasattr(self, 'fc_src'):
                    feat_src = self.fc(h_src).view(
                        *src_prefix_shape, self._num_heads, self._out_feats)
                    feat_dst = self.fc(h_dst).view(
                        *dst_prefix_shape, self._num_heads, self._out_feats)
                else:
                    feat_src = self.fc_src(h_src).view(
                        *src_prefix_shape, self._num_heads, self._out_feats)
                    feat_dst = self.fc_dst(h_dst).view(
                        *dst_prefix_shape, self._num_heads, self._out_feats)
            else:
                src_prefix_shape = dst_prefix_shape = feat.shape[:-1]
                h_src = h_dst = self.feat_drop(feat)
                feat_src = feat_dst = self.fc(h_src).view(
                    *src_prefix_shape, self._num_heads, self._out_feats)
                if graph.is_block:
                    feat_dst = feat_src[:graph.number_of_dst_nodes()]
                    h_dst = h_dst[:graph.number_of_dst_nodes()]
                    dst_prefix_shape = (graph.number_of_dst_nodes(),) + dst_prefix_shape[1:]
            # NOTE: GAT paper uses "first concatenation then linear projection"
            # to compute attention scores, while ours is "first projection then
            # addition", the two approaches are mathematically equivalent:
            # We decompose the weight vector a mentioned in the paper into
            # [a_l || a_r], then
            # a^T [Wh_i || Wh_j] = a_l Wh_i + a_r Wh_j
            # Our implementation is much efficient because we do not need to
            # save [Wh_i || Wh_j] on edges, which is not memory-efficient. Plus,
            # addition could be optimized with DGL's built-in function u_add_v,
            # which further speeds up computation and saves memory footprint.
            el = (feat_src * self.attn_l).sum(dim=-1).unsqueeze(-1)
            er = (feat_dst * self.attn_r).sum(dim=-1).unsqueeze(-1)
            graph.srcdata.update({'ft': feat_src, 'el': el})
            graph.dstdata.update({'er': er})
            # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
            graph.apply_edges(fn.u_add_v('el', 'er', 'e'))
            e = self.leaky_relu(graph.edata.pop('e'))
            # e[e == 0] = -1e3
            # e = graph.edata.pop('e')
            # compute softmax
            graph.edata['a'] = self.attn_drop(edge_softmax(graph, e))
            # message passing
            graph.update_all(fn.u_mul_e('ft', 'a', 'm'),
                             fn.sum('m', 'ft'))
            rst = graph.dstdata['ft']

            # bias
            if self.bias is not None:
                rst = rst + self.bias.view(
                    *((1,) * len(dst_prefix_shape)), self._num_heads, self._out_feats)

            # residual
            if self.res_fc is not None:
                # Use -1 rather than self._num_heads to handle broadcasting
                resval = self.res_fc(h_dst).view(*dst_prefix_shape, -1, self._out_feats)
                rst = rst + resval

            if self._concat_out:
                rst = rst.flatten(1)
            else:
                rst = torch.mean(rst, dim=1)

            if self.norm is not None:
                rst = self.norm(rst)

            # activation
            if self.activation:
                rst = self.activation(rst)

            if get_attention:
                return rst, graph.edata['a']
            else:
                return rst



import pandas as pd
import numpy as np

class parametersDict(object):
    def __init__(self,parameters_dict):
        self.parameters_dict = parameters_dict
        self.length = []
        self.keys = []
        for key,values in self.parameters_dict.items():
            self.keys.append(key)
            self.length.append(len(values))
        self.nums = [1] * len(self.length)
        for i in range(len(self.length)):
            for j in range(i,len(self.length)):
                self.nums[i] *= self.length[j]
        self.para_dis = []
        print(self.length)
        print(self.nums)
                
    def getindex(self,index):
        result = []
        value = index
        for i in range(len(self.nums) - 1):
            result.append(value // self.nums[i+1])
            value = value - result[i] * self.nums[i+1]
        result.append(value) 
        result_dict = dict()
        for index,value in enumerate(result):
            result_dict[self.keys[index]] = self.parameters_dict.get(self.keys[index])[value]
        return result_dict
    
    def myiter(self):
        for i in range(0,self.nums[0]):
            self.para_dis.append(self.getindex(i))
        return self.para_dis
